# Copyright (c)  2021  Xiaomi Corporation (authors: Fangjun Kuang)

from typing import List, Optional, Union

import _kaldifeat
import torch
import torch.nn as nn


class OfflineFeature(nn.Module):
    """Offline feature is a base class of other feature computers,
    e.g., Fbank, Mfcc.

    This class has two fields:

        (1) opts. It contains the options for the feature computer.
        (2) computer. The actual feature computer. It should be
            instantiated by subclasses.
    """

    def __init__(self, opts):
        super().__init__()

        self.opts = opts

        # self.computer is expected to be set by subclasses
        self.computer = None

    def forward(
        self,
        waves: Union[torch.Tensor, List[torch.Tensor]],
        vtln_warp: float = 1.0,
        chunk_size: Optional[int] = None,
    ) -> Union[torch.Tensor, List[torch.Tensor]]:
        """Compute the features of a single waveform or
        a list of waveforms.

        Args:
          waves:
            A single 1-D tensor or a  list of 1-D tensors. Each tensor contains
            audio samples of a sound file. To get a result compatible with
            Kaldi, you should scale the samples to [-32768, 32767] before
            calling this function. Note: You are not required to scale them if
            you don't care about the compatibility with Kaldi.
          vtln_warp
            The VTLN warping factor that the user wants to be applied when
            computing features for this utterance.  Will normally be 1.0,
            meaning no warping is to be done.  The value will be ignored for
            feature types that don't support VLTN, such as spectrogram features.
          chunk_size:
            It specifies the number of frames for each computation.
            If None, it computes features at once (requiring more memory for
            long waves). If not None, each computation takes this number of
            frames (requiring less memory).
        Returns:
          Return a list of 2-D tensors containing the features if the
          input is a list of 1-D tensors. The returned list has as many elements
          as the input list.
          Return a single 2-D tensor if the input is a single tensor.
        """
        if isinstance(waves, list):
            is_list = True
        else:
            waves = [waves]
            is_list = False

        num_frames_per_wave = [
            _kaldifeat.num_frames(w.numel(), self.opts.frame_opts)
            for w in waves
        ]

        strided = [self.convert_samples_to_frames(w) for w in waves]
        strided = torch.cat(strided, dim=0)

        features = self.compute(strided, vtln_warp)

        if is_list:
            return list(features.split(num_frames_per_wave))
        else:
            return features

    def compute(
        self,
        x: torch.Tensor,
        vtln_warp: float = 1.0,
        chunk_size: Optional[int] = None,
    ) -> torch.Tensor:
        """Compute features given a 2-D tensor containing
        frames data. Each row is a frame of size frame_lens, specified
        in the options.
        Args:
          x:
            A 2-D tensor.
          vtln_warp
            The VTLN warping factor that the user wants to be applied when
            computing features for this utterance.  Will normally be 1.0,
            meaning no warping is to be done.  The value will be ignored for
            feature types that don't support VLTN, such as spectrogram features.
          chunk_size:
            It specifies the number of frames for each computation.
            If None, it computes features at once (requiring more memory for
            long waves). If not None, each computation takes this number of
            frames (requiring less memory).
        Returns:
          Return a 2-D tensor with as many rows as the input tensor. Its
          number of columns is the number mel bins.
        """
        assert x.ndim == 2
        assert x.dtype == torch.float32
        if chunk_size is None:
            features = self.computer.compute_features(x, vtln_warp)
        else:
            assert chunk_size > 0
            num_chunks = x.size(0) // chunk_size
            end = 0
            features = []
            for i in range(num_chunks):
                start = i * chunk_size
                end = start + chunk_size
                this_chunk = self.computer.compute_features(
                    x[start:end], vtln_warp
                )
                features.append(this_chunk)
            if end < x.size(0):
                last_chunk = self.compute_features(x[end:], vtln_warp)
                features.append(last_chunk)
            features = torch.cat(features, dim=0)

        return features

    def convert_samples_to_frames(self, wave: torch.Tensor) -> torch.Tensor:
        """Convert a 1-D tensor containing audio samples to a 2-D
        tensor where each row is a frame of samples of size frame length
        specified in the options.

        Args:
          waves:
            A 1-D tensor.
        Returns:
          Return a 2-D tensor.
        """
        assert wave.ndim == 1
        assert wave.dtype == torch.float32
        return _kaldifeat.get_strided(wave, self.opts.frame_opts)



# Copyright      2021  Xiaomi Corporation (authors: Fangjun Kuang)

from typing import List

import numpy as np
import soundfile as sf
import torch


def read_wave(filename) -> torch.Tensor:
    """Read a wave file and return it as a 1-D tensor.

    Note:
      You don't need to scale it to [-32768, 32767].
      We use scaling here to follow the approach in Kaldi.

    Args:
      filename:
        Filename of a sound file.
    Returns:
      Return a 1-D tensor containing audio samples.
    """
    with sf.SoundFile(filename) as sf_desc:
        sampling_rate = sf_desc.samplerate
        assert sampling_rate == 16000
        data = sf_desc.read(dtype=np.float32, always_2d=False)
    data *= 32768
    return torch.from_numpy(data)


def read_ark_txt(filename) -> torch.Tensor:
    #  test_data_dir = cur_dir / "test_data"
    #  filename = test_data_dir / filename
    features = []
    with open(filename) as f:
        for line in f:
            if "[" in line:
                continue
            line = line.strip("").split()
            data = [float(d) for d in line if d != "]"]
            features.append(data)
    ans = torch.tensor(features)
    return ans


def get_devices() -> List[torch.device]:
    ans = [torch.device("cpu")]
    if torch.cuda.is_available():
        ans.append(torch.device("cuda", 0))
        if torch.cuda.device_count() > 1:
            ans.append(torch.device("cuda", 1))
    return ans



#!/usr/bin/env python3
#
# Copyright (c)  2021  Xiaomi Corporation (authors: Fangjun Kuang)

import sys
from pathlib import Path

cur_dir = Path(__file__).resolve().parent
kaldi_feat_dir = cur_dir.parent.parent.parent


import torch

sys.path.insert(0, f"{kaldi_feat_dir}/build/lib")

import kaldifeat


def test_frame_extraction_options():
    opts = kaldifeat.FrameExtractionOptions()
    opts.samp_freq = 220500
    opts.frame_shift_ms = 15
    opts.frame_length_ms = 40
    opts.dither = 0.1
    opts.preemph_coeff = 0.98
    opts.remove_dc_offset = False
    opts.window_type = "hanning"
    opts.round_to_power_of_two = False
    opts.blackman_coeff = 0.422
    opts.snip_edges = False
    print(opts)


def test_mel_banks_options():
    opts = kaldifeat.MelBanksOptions()
    opts.num_bins = 23
    opts.low_freq = 21
    opts.high_freq = 8000
    opts.vtln_low = 101
    opts.vtln_high = -501
    opts.debug_mel = True
    opts.htk_mode = True
    print(opts)


def test_fbank_options():
    opts = kaldifeat.FbankOptions()
    frame_opts = opts.frame_opts
    mel_opts = opts.mel_opts

    opts.energy_floor = 0
    opts.htk_compat = False
    opts.raw_energy = True
    opts.use_energy = False
    opts.use_log_fbank = True
    opts.use_power = True
    opts.device = torch.device("cuda", 0)

    frame_opts.blackman_coeff = 0.42
    frame_opts.dither = 1
    frame_opts.frame_length_ms = 25
    frame_opts.frame_shift_ms = 10
    frame_opts.preemph_coeff = 0.97
    frame_opts.remove_dc_offset = True
    frame_opts.round_to_power_of_two = True
    frame_opts.samp_freq = 16000
    frame_opts.snip_edges = True
    frame_opts.window_type = "povey"

    mel_opts.debug_mel = True
    mel_opts.high_freq = 0
    mel_opts.low_freq = 20
    mel_opts.num_bins = 23
    mel_opts.vtln_high = -500
    mel_opts.vtln_low = 100

    print(opts)


def test_mfcc_options():
    opts = kaldifeat.MfccOptions()
    frame_opts = opts.frame_opts
    mel_opts = opts.mel_opts

    opts.num_ceps = 10
    opts.use_energy = False
    opts.energy_floor = 0.0
    opts.raw_energy = True
    opts.cepstral_lifter = 22.0
    opts.htk_compat = False
    opts.device = torch.device("cpu")

    frame_opts.blackman_coeff = 0.42
    frame_opts.dither = 1
    frame_opts.frame_length_ms = 25
    frame_opts.frame_shift_ms = 10
    frame_opts.preemph_coeff = 0.97
    frame_opts.remove_dc_offset = True
    frame_opts.round_to_power_of_two = True
    frame_opts.samp_freq = 16000
    frame_opts.snip_edges = True
    frame_opts.window_type = "povey"

    mel_opts.debug_mel = True
    mel_opts.high_freq = 0
    mel_opts.low_freq = 20
    mel_opts.num_bins = 23
    mel_opts.vtln_high = -500
    mel_opts.vtln_low = 100

    print(opts)


def test_spectogram_options():
    opts = kaldifeat.SpectrogramOptions()
    opts.energy_floor = 0.0
    opts.raw_energy = True

    frame_opts = opts.frame_opts
    frame_opts.blackman_coeff = 0.42
    frame_opts.dither = 1
    frame_opts.frame_length_ms = 25
    frame_opts.frame_shift_ms = 10
    frame_opts.preemph_coeff = 0.97
    frame_opts.remove_dc_offset = True
    frame_opts.round_to_power_of_two = True
    frame_opts.samp_freq = 16000
    frame_opts.snip_edges = True
    frame_opts.window_type = "povey"

    print(opts)


def test_plp_options():
    opts = kaldifeat.PlpOptions()
    opts.lpc_order = 12
    opts.num_ceps = 13
    opts.use_energy = True
    opts.energy_floor = 0.0
    opts.raw_energy = True
    opts.compress_factor = 0.33333
    opts.cepstral_lifter = 22
    opts.cepstral_scale = 1.0
    opts.htk_compat = False
    opts.device = torch.device("cpu")

    frame_opts = opts.frame_opts
    frame_opts.blackman_coeff = 0.42
    frame_opts.dither = 1
    frame_opts.frame_length_ms = 25
    frame_opts.frame_shift_ms = 10
    frame_opts.preemph_coeff = 0.97
    frame_opts.remove_dc_offset = True
    frame_opts.round_to_power_of_two = True
    frame_opts.samp_freq = 16000
    frame_opts.snip_edges = True
    frame_opts.window_type = "povey"

    mel_opts = opts.mel_opts
    mel_opts.debug_mel = True
    mel_opts.high_freq = 0
    mel_opts.low_freq = 20
    mel_opts.num_bins = 23
    mel_opts.vtln_high = -500
    mel_opts.vtln_low = 100

    print(opts)


def main():
    test_frame_extraction_options()
    test_mel_banks_options()
    test_fbank_options()
    test_mfcc_options()
    test_spectogram_options()
    test_plp_options()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright      2021  Xiaomi Corporation (authors: Fangjun Kuang)

from pathlib import Path

import torch
from utils import get_devices, read_ark_txt, read_wave

import kaldifeat

cur_dir = Path(__file__).resolve().parent


def test_fbank_default():
    print("=====test_fbank_default=====")
    for device in get_devices():
        print("device", device)
        opts = kaldifeat.FbankOptions()
        opts.device = device
        opts.frame_opts.dither = 0
        fbank = kaldifeat.Fbank(opts)
        filename = cur_dir / "test_data/test.wav"
        wave = read_wave(filename).to(device)

        features = fbank(wave)
        gt = read_ark_txt(cur_dir / "test_data/test.txt")
        assert torch.allclose(features.cpu(), gt, rtol=1e-1)


def test_fbank_htk():
    print("=====test_fbank_htk=====")
    for device in get_devices():
        print("device", device)
        opts = kaldifeat.FbankOptions()
        opts.device = device
        opts.frame_opts.dither = 0
        opts.use_energy = True
        opts.htk_compat = True

        fbank = kaldifeat.Fbank(opts)
        filename = cur_dir / "test_data/test.wav"
        wave = read_wave(filename).to(device)

        features = fbank(wave)
        gt = read_ark_txt(cur_dir / "test_data/test-htk.txt")
        assert torch.allclose(features.cpu(), gt, rtol=1e-1)


def test_fbank_with_energy():
    print("=====test_fbank_with_energy=====")
    for device in get_devices():
        print("device", device)
        opts = kaldifeat.FbankOptions()
        opts.device = device
        opts.frame_opts.dither = 0
        opts.use_energy = True

        fbank = kaldifeat.Fbank(opts)
        filename = cur_dir / "test_data/test.wav"
        wave = read_wave(filename).to(device)

        features = fbank(wave)
        gt = read_ark_txt(cur_dir / "test_data/test-with-energy.txt")
        assert torch.allclose(features.cpu(), gt, rtol=1e-1)


def test_fbank_40_bins():
    print("=====test_fbank_40_bins=====")
    for device in get_devices():
        print("device", device)
        opts = kaldifeat.FbankOptions()
        opts.device = device
        opts.frame_opts.dither = 0
        opts.mel_opts.num_bins = 40

        fbank = kaldifeat.Fbank(opts)
        filename = cur_dir / "test_data/test.wav"
        wave = read_wave(filename).to(device)

        features = fbank(wave)
        gt = read_ark_txt(cur_dir / "test_data/test-40.txt")
        assert torch.allclose(features.cpu(), gt, rtol=1e-1)


def test_fbank_40_bins_no_snip_edges():
    print("=====test_fbank_40_bins_no_snip_edges=====")
    for device in get_devices():
        print("device", device)
        opts = kaldifeat.FbankOptions()
        opts.device = device
        opts.frame_opts.dither = 0
        opts.mel_opts.num_bins = 40
        opts.frame_opts.snip_edges = False

        fbank = kaldifeat.Fbank(opts)
        filename = cur_dir / "test_data/test.wav"
        wave = read_wave(filename).to(device)

        features = fbank(wave)
        gt = read_ark_txt(cur_dir / "test_data/test-40-no-snip-edges.txt")
        assert torch.allclose(features.cpu(), gt, rtol=1e-1)


def test_fbank_chunk():
    print("=====test_fbank_chunk=====")

    filename = cur_dir / "test_data/test-1hour.wav"
    if filename.is_file() is False:
        print(
            f"Please execute {cur_dir}/test_data/run.sh "
            f"to generate {filename} before running tis test"
        )
        return

    for device in get_devices():
        print("device", device)
        opts = kaldifeat.FbankOptions()
        opts.device = device
        opts.frame_opts.dither = 0
        opts.mel_opts.num_bins = 40
        opts.frame_opts.snip_edges = False

        fbank = kaldifeat.Fbank(opts)
        wave = read_wave(filename).to(device)

        # You can use
        #
        #  $ watch -n 0.2 free -m
        #
        # to view memory consumption
        #
        # 100 frames per chunk
        features = fbank(wave, chunk_size=100 * 10)
        print(features.shape)


def test_fbank_batch():
    print("=====test_fbank_chunk=====")
    for device in get_devices():
        print("device", device)
        wave0 = read_wave(cur_dir / "test_data/test.wav").to(device)
        wave1 = read_wave(cur_dir / "test_data/test2.wav").to(device)

        opts = kaldifeat.FbankOptions()
        opts.device = device
        opts.frame_opts.dither = 0
        fbank = kaldifeat.Fbank(opts)

        features = fbank([wave0, wave1], chunk_size=10)

        features0 = fbank(wave0)
        features1 = fbank(wave1)

        assert torch.allclose(features[0], features0)
        assert torch.allclose(features[1], features1)


if __name__ == "__main__":
    test_fbank_default()
    test_fbank_htk()
    test_fbank_with_energy()
    test_fbank_40_bins()
    test_fbank_40_bins_no_snip_edges()
    test_fbank_chunk()
    test_fbank_batch()



#!/usr/bin/env python3

# Copyright      2021  Xiaomi Corporation (authors: Fangjun Kuang)

from pathlib import Path

import torch
from utils import get_devices, read_ark_txt, read_wave

import kaldifeat

cur_dir = Path(__file__).resolve().parent


def test_mfcc_default():
    print("=====test_mfcc_default=====")
    for device in get_devices():
        print("device", device)
        opts = kaldifeat.MfccOptions()
        opts.device = device
        opts.frame_opts.dither = 0
        mfcc = kaldifeat.Mfcc(opts)
        filename = cur_dir / "test_data/test.wav"
        wave = read_wave(filename).to(device)

        features = mfcc(wave)
        gt = read_ark_txt(cur_dir / "test_data/test-mfcc.txt")
        assert torch.allclose(features.cpu(), gt, atol=1e-1)


def test_mfcc_no_snip_edges():
    print("=====test_mfcc_no_snip_edges=====")
    for device in get_devices():
        print("device", device)
        opts = kaldifeat.MfccOptions()
        opts.device = device
        opts.frame_opts.dither = 0
        opts.frame_opts.snip_edges = False

        mfcc = kaldifeat.Mfcc(opts)
        filename = cur_dir / "test_data/test.wav"
        wave = read_wave(filename).to(device)

        features = mfcc(wave)
        gt = read_ark_txt(cur_dir / "test_data/test-mfcc-no-snip-edges.txt")
        assert torch.allclose(features.cpu(), gt, rtol=1e-1)


if __name__ == "__main__":
    test_mfcc_default()
    test_mfcc_no_snip_edges()



# Copyright (c)  2021  Xiaomi Corporation (authors: Fangjun Kuang)


import _kaldifeat

from .offline_feature import OfflineFeature


class Plp(OfflineFeature):
    def __init__(self, opts: _kaldifeat.PlpOptions):
        super().__init__(opts)
        self.computer = _kaldifeat.Plp(opts)



import torch
from _kaldifeat import (
    FbankOptions,
    FrameExtractionOptions,
    MelBanksOptions,
    MfccOptions,
    PlpOptions,
    SpectrogramOptions,
)

from .fbank import Fbank
from .mfcc import Mfcc
from .plp import Plp
from .spectrogram import Spectrogram
__version__ = '1.5.3.dev0'



# Copyright (c)  2021  Xiaomi Corporation (author: Fangjun Kuang)

import glob
import os
import platform
import shutil
import sys
from pathlib import Path

import setuptools
from setuptools.command.build_ext import build_ext


def is_for_pypi():
    ans = os.environ.get("KALDIFEAT_IS_FOR_PYPI", None)
    return ans is not None


def is_macos():
    return platform.system() == "Darwin"


try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            # In this case, the generated wheel has a name in the form
            # k2-xxx-pyxx-none-any.whl
            if is_for_pypi() and not is_macos():
                self.root_is_pure = True
            else:
                # The generated wheel has a name ending with
                # -linux_x86_64.whl
                self.root_is_pure = False


except ImportError:
    bdist_wheel = None


def cmake_extension(name, *args, **kwargs) -> setuptools.Extension:
    kwargs["language"] = "c++"
    sources = []
    return setuptools.Extension(name, sources, *args, **kwargs)


class BuildExtension(build_ext):
    def build_extension(self, ext: setuptools.extension.Extension):
        # build/temp.linux-x86_64-3.8
        os.makedirs(self.build_temp, exist_ok=True)

        # build/lib.linux-x86_64-3.8
        os.makedirs(self.build_lib, exist_ok=True)

        kaldifeat_dir = Path(__file__).parent.parent.resolve()

        cmake_args = os.environ.get("KALDIFEAT_CMAKE_ARGS", "")
        make_args = os.environ.get("KALDIFEAT_MAKE_ARGS", "")
        system_make_args = os.environ.get("MAKEFLAGS", "")

        if cmake_args == "":
            cmake_args = "-DCMAKE_BUILD_TYPE=Release"

        if make_args == "" and system_make_args == "":
            print("For fast compilation, run:")
            print('export KALDIFEAT_MAKE_ARGS="-j"; python setup.py install')

        if "PYTHON_EXECUTABLE" not in cmake_args:
            print(f"Setting PYTHON_EXECUTABLE to {sys.executable}")
            cmake_args += f" -DPYTHON_EXECUTABLE={sys.executable}"

        build_cmd = f"""
            cd {self.build_temp}

            cmake {cmake_args} {kaldifeat_dir}


            make {make_args} _kaldifeat
        """
        print(f"build command is:\n{build_cmd}")

        ret = os.system(build_cmd)
        if ret != 0:
            raise Exception(
                "\nBuild kaldifeat failed. Please check the error message.\n"
                "You can ask for help by creating an issue on GitHub.\n"
                "\nClick:\n\thttps://github.com/csukuangfj/kaldifeat/issues/new\n"  # noqa
            )

        lib_so = glob.glob(f"{self.build_temp}/lib/*kaldifeat*.so")
        lib_so += glob.glob(f"{self.build_temp}/lib/*kaldifeat*.dylib")  # macOS
        for so in lib_so:
            print(f"Copying {so} to {self.build_lib}/")
            shutil.copy(f"{so}", f"{self.build_lib}/")



#!/usr/bin/env python3
#
# Copyright (c)  2021  Xiaomi Corporation (author: Fangjun Kuang)

import re

import setuptools

from cmake.cmake_extension import BuildExtension, bdist_wheel, cmake_extension


def read_long_description():
    with open("README.md", encoding="utf8") as f:
        readme = f.read()
    return readme


def get_package_version():
    with open("CMakeLists.txt") as f:
        content = f.read()

    match = re.search(r"set\(kaldifeat_VERSION (.*)\)", content)
    latest_version = match.group(1).strip('"')
    return latest_version


package_name = "kaldifeat"

with open("kaldifeat/python/kaldifeat/__init__.py", "a") as f:
    f.write(f"__version__ = '{get_package_version()}'\n")

setuptools.setup(
    name=package_name,
    version=get_package_version(),
    author="Fangjun Kuang",
    author_email="csukuangfj@gmail.com",
    data_files=[("", ["LICENSE", "README.md"])],
    package_dir={package_name: "kaldifeat/python/kaldifeat"},
    packages=[package_name],
    url="https://github.com/csukuangfj/kaldifeat",
    long_description=read_long_description(),
    long_description_content_type="text/markdown",
    ext_modules=[cmake_extension("_kaldifeat")],
    cmdclass={"build_ext": BuildExtension, "bdist_wheel": bdist_wheel},
    classifiers=[
        "Programming Language :: C++",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.6.0",
    license="Apache licensed, as found in the LICENSE file",
)

# remove the line __version__ from kaldifeat/python/kaldifeat/__init__.py
with open("kaldifeat/python/kaldifeat/__init__.py", "r") as f:
    lines = f.readlines()

with open("kaldifeat/python/kaldifeat/__init__.py", "w") as f:
    for line in lines:
        if "__version__" not in line:
            f.write(line)
